     ////////////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM         //
   // Written by Andrew Broad                     ///
  //                                             ////
 // DefinedTypeInstance.java (1998:10:27 11:08) /////
/////////////////////////////////////////////////////
// Methods of DefinedTypeInstance:              /////
// - DefinedTypeInstance constructor            /////
// - getType                                    ////
// - getValue                                   ///
// - toString                                   //
/////////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `DefinedTypeInstance' is used to wrap an instance of a defined
** type.
*******************************************************************************/

public class DefinedTypeInstance implements Serializable {
	private DefinedType _type;
	private Object _value;

	/**********************************************************************/
	// DefinedTypeInstance constructor : (DefinedType x Object) ->
	/***********************************************************************
	** Constructs this defined type instance by plugging in the given value.
	** This method is called from the constructors of subclasses of
	** DefinedTypeInstance.
	** @param type The defined type of which this is an instance.
	** @param value The value of this defined type instance, which must
	**  conform to the underlying type.
	***********************************************************************/
 
 	public DefinedTypeInstance (DefinedType type, Object value)
	                           /* throws InvalidTypeValue */
	{
		_type= type;
		_value= value;
		//type.validateValue(value);
	}
	
	/**********************************************************************/
	// DefinedTypeInstance.getType : void -> Type
	/***********************************************************************
	** Returns the defined type of which this object is an instance.
	** @return The defined type of which this is an instance.
	***********************************************************************/

	public DefinedType getType () {
		return _type;
	}

	/**********************************************************************/
	// DefinedTypeInstance.getValue : void -> Object
	/***********************************************************************
	** Returns the value of this defined type instance.
	** @return The value of this defined type instance.
	***********************************************************************/

	public Object getValue () {
		return _value;
	}

	/**********************************************************************/
	// DefinedTypeInstance.toString : void -> String
	/***********************************************************************
	** Returns a string representing this defined type instance.
	** @returns The string representation of this defined type instance.
	***********************************************************************/

	public String toString () {
		return _value.toString();
	}
}
