     ////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM //
   // Written by Andrew Broad             ///
  //                                     ////
 // Frame.java (1998:11:03 15:16)       /////
/////////////////////////////////////////////
// Methods of Frame:                    /////
// - static initialise                  /////
// - Frame constructor                  /////
// - getFrameClass                      /////
// - getNumber                          /////
// - setNumber                          /////
// - getSlotFiller(Slot)                /////
// - getSlotFiller(String)              /////
// - setSlotFiller(Slot,Object)         /////
// - setSlotFillerIndirectly            /////
// - setSlotFiller(String,Object)       /////
// - isSlotSetDirectly                  /////
// - getSlots                           /////
// - markCaseProblemPart                /////
// - isCaseProblemPart                  /////
// - toString                           /////
// - print                              /////
// - static generateFIL                 /////
// - generateFIL                        ////
// - save                               ///
// - static load                        //
/////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `Frame' represents a frame (an instance of a
** <a href="FrameClass.html">FrameClass</a>).
*******************************************************************************/

public class Frame implements Serializable {
	private Hashtable/*Slot->Object*/ _frameRep;
	private Hashtable/*Slot->boolean*/ _slotsSetDirectly;
	private FrameClass _frameClass;
	protected int _number;
	boolean _caseProblemPart= false;

	/**********************************************************************/
	// static Frame.initialise : void -> void
	/***********************************************************************
	** Initialises the exjava subclasses of Frame, after loading the
	** EXPRESS metamodel from EXPRESS/frame_classes.exp.
	***********************************************************************/

	public static void initialise () {
		Express metamodel= new Express("EXPRESS/frame_classes.exp");

		DefinedTypeFrame.initialise(metamodel);
		EntityFrame.initialise(metamodel);
		ExplicitAttrFrame.initialise(metamodel);
		InverseAttrFrame.initialise(metamodel);
		DerivedAttrFrame.initialise(metamodel);
		DomainRuleFrame.initialise(metamodel);
		UniqueRuleFrame.initialise(metamodel);
		LiteralFrame.initialise(metamodel);
		BuiltInFrame.initialise(metamodel);
		UnaryOpFrame.initialise(metamodel);
		BinaryOpFrame.initialise(metamodel);
		QueryFrame.initialise(metamodel);
		BuiltInFnCallFrame.initialise(metamodel);
		UserFnCallFrame.initialise(metamodel);
		QualifierFrame.initialise(metamodel);
		AggInitFrame.initialise(metamodel);
		AggInitElemFrame.initialise(metamodel);
		AggTypeFrame.initialise(metamodel);
		OneofFrame.initialise(metamodel);
		AndFrame.initialise(metamodel);
		AndorFrame.initialise(metamodel);

		Schema[] schemata= metamodel.schemas;
		for (int s= 0; s < schemata.length; s++) {
			Type[] types= schemata[s].types;
			for (int t= 0; t < types.length; t++)
				new DefinedType(types[t]);

			Entity[] entities= schemata[s].entities;
			for (int e= 0; e < entities.length; e++)
				try {
					FrameClass.lookup(entities[e].id,
					                  metamodel);
				} catch (FrameClassDoesNotExist x) {
					System.err.println("It's all gone "
					+ "horribly wrong!");
					System.exit(1);
				}
	       	}

		DefinedTypeFrame.buildQueuedFrames();
		EntityFrame.buildQueuedFrames();
		ExplicitAttrFrame.buildQueuedFrames();
		InverseAttrFrame.buildQueuedFrames();
		DerivedAttrFrame.buildQueuedFrames();
		DomainRuleFrame.buildQueuedFrames();
		UniqueRuleFrame.buildQueuedFrames();
		LiteralFrame.buildQueuedFrames();
		BuiltInFrame.buildQueuedFrames();
		UnaryOpFrame.buildQueuedFrames();
		BinaryOpFrame.buildQueuedFrames();
		QueryFrame.buildQueuedFrames();
		BuiltInFnCallFrame.buildQueuedFrames();
		UserFnCallFrame.buildQueuedFrames();
		QualifierFrame.buildQueuedFrames();
		AggInitFrame.buildQueuedFrames();
		AggInitElemFrame.buildQueuedFrames();
		AggTypeFrame.buildQueuedFrames();
		OneofFrame.buildQueuedFrames();
		AndFrame.buildQueuedFrames();
		AndorFrame.buildQueuedFrames();
	}

	/**********************************************************************/
	// Frame constructor : Hashtable ->
	/***********************************************************************
	** Constructs this frame. This method is called from the constructors of
	** subclasses of Frame. It should be called with "new Hashtable" for
	** instanceData.
	** @param instanceData The initial slot fillers for the frame in a hash
	**  table which associates <a href="Slot.html">Slot</a> objects with
	**  their filler values. The use of this parameter with any actual
	**  parameter other than "new Hashtable" is <b>deprecated</b> -
	**  construct the frame with a blank hash table instead, and use
	**  setSlotFiller to fill the slots <i>after</i> constructing the frame.
	***********************************************************************/

	protected Frame (Hashtable instanceData) {
		_slotsSetDirectly= new Hashtable();
		_frameRep= instanceData;
	}

	/**********************************************************************/
	// Frame constructor : (FrameClass x Hashtable) ->
	/***********************************************************************
	** Constructs this frame.
	** @param frameClass The frame class of the frame to be constructed,
	**  which specifies the set of slots the frame should have.
	** @param instanceData The initial slot fillers for the frame in a hash
	**  table which associates <a href="Slot.html">Slot</a> objects with
	**  their filler values. The use of this parameter with any actual
	**  parameter other than "new Hashtable" is <b>deprecated</b> -
	**  construct the frame with a blank hash table instead, and use
	**  setSlotFiller to fill the slots <i>after</i> constructing the frame.
	** @exception InvalidSlot If `instanceData' includes slots other than
	**  those specified in `frameClass'.
	***********************************************************************/
 
 	public Frame (FrameClass frameClass, Hashtable instanceData)
	             throws InvalidSlot
	{
		_slotsSetDirectly= new Hashtable();
		_frameClass= frameClass;
		try {
			frameClass.addInstance(this);
		} catch (InvalidFrameClass x) {
			System.err.println("Something has gone badly wrong for "
			+ "an InvalidFrameClass to be thrown here!");
		}
		_frameRep= instanceData;
		frameClass.validateFrame(this);
		Enumeration slots= _frameRep.keys();
		while (slots.hasMoreElements()) {
			if (! frameClass.getSlots().contains
				((Slot)slots.nextElement())) {
				throw new InvalidSlot();
			}
		}
		_number= frameClass.getInstances().size();
	}
	
	/**********************************************************************/
	// Frame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}

	/**********************************************************************/
	// Frame.getNumber : void -> int
	/***********************************************************************
	** Returns this frame's instance number (unique to the frame class).
	** @returns This frame's instance number.
	***********************************************************************/

	public int getNumber () {
		return _number;
	}
		
	/**********************************************************************/
	// Frame.setNumber : int -> void
	/***********************************************************************
	** Gives this frame an instance number which is unique to this frame
	** class.
	** @param An instance number for this frame.
	***********************************************************************/

	public void setNumber (int number) {
		_number= number;
	}
		
	/**********************************************************************/
	// Frame.getSlotFiller : Slot -> Object
	/***********************************************************************
	** Returns the filler of a given slot.
	** @param slot The slot whose filler you want to read.
	** @return The filler of `slot', as held in this frame.
	***********************************************************************/
 
 	public Object getSlotFiller (Slot slot) {
		return _frameRep.get(slot);
	}

	/**********************************************************************/
	// Frame.getSlotFiller : String -> Object
	/***********************************************************************
	** Returns the filler of a given slot.
	** @param slot The name of the slot whose filler you want to read.
	** @return The filler of the slot, as held in this frame.
	** @exception InvalidSlot If the frame does not have a slot of that name
	**  (as specified in its frame class).
	***********************************************************************/

	public Object getSlotFiller (String slotName)
	                            throws InvalidSlot
	{
		return getSlotFiller(getFrameClass().getSlot(slotName));
	}

	/**********************************************************************/
	// Frame.setSlotFiller : (Slot x Object) -> void
	/***********************************************************************
	** Fills a given slot with a given value. Records this slot as having
	** been written directly.
	** @param slot The slot you want to write into.
	** @param filler The filler (value) you want to write into `slot',
	**               which must be of the right type for `slot'.
	***********************************************************************/
 
	public void setSlotFiller (Slot slot, Object filler)
	                          /* throws InvalidSlotFiller */
	{
// if (slot.validFiller(filler))

		_slotsSetDirectly.put(slot,new Boolean(true));

		if (filler != null)
			_frameRep.put(slot,filler);
		else
			_frameRep.remove(slot);

		if (filler instanceof Frame) {
			setInverseSlot(slot,(Frame)filler);
			if (_caseProblemPart)
				((Frame)filler).markCaseProblemPart();
		}
		else if (filler instanceof Vector) {
			Enumeration fillers= ((Vector)filler).elements();
			while (fillers.hasMoreElements()) {
				Object nextFiller= fillers.nextElement();
				if (nextFiller instanceof Frame) {
					setInverseSlot(slot,(Frame)nextFiller);
					if (_caseProblemPart)
						((Frame)nextFiller).
							markCaseProblemPart();
				}
				else break;
			}
		}

// else throw new InvalidSlotFiller();
	}

	/**********************************************************************/
	// Frame.setInverseSlot : (Slot x Frame) -> void
	/***********************************************************************
	** If the given slot has an inverse slot, then set that inverse slot
	** to this frame for the given filler frame.
	** @param slot The slot being set.
	** @param filler The filler of the slot being set (a frame).
	** @param filler The filler (value) you want to write into `slot',
	**               which must be of the right type for `slot'.
	***********************************************************************/
 
	private void setInverseSlot (Slot slot, Frame filler) {
		Slot inverseSlot= Slot.lookupInverseSlot
		                      (slot.getFrameClassDeclaredIn().getName(),
		                       slot.getName(),
		                       filler.getFrameClass());

		if (inverseSlot != null) {
			if (filler.isSlotSetDirectly(inverseSlot))
				return;

			if (! inverseSlot.isAggType())
				filler.setSlotFillerIndirectly
				       (inverseSlot,this);
			else {
				Vector agg= (Vector)filler.
				                    getSlotFiller(inverseSlot);
				if (agg == null) {
					String kind= null;
					try {
						kind= ((Frame)inverseSlot.
							getType()).
							getSlotFiller("kind").
							toString();
					} catch (InvalidSlot x) {
						System.err.println("Slot type "
						+ "frame doesn't have a `kind' "
						+ "slot!");
						System.exit(1);
					}
					if (kind.equals("list_type") ||
					    kind.equals("array_type"))
						agg= new OrderedVector();
					else if (kind.equals("set_type") ||
					         kind.equals("bag_type"))
						agg= new UnorderedVector();
					else {
						System.err.println("Frame.java:"
						+ ":311: `kind' has the follow"+
						"ing anomalous value: " + kind);
						System.exit(1);
					}
				}
				agg.addElement(this);
				filler.setSlotFillerIndirectly(inverseSlot,agg);
			}
		}
	}

	/**********************************************************************/
	// Frame.setSlotFillerIndirectly : (Slot x Object) -> void
	/***********************************************************************
	** Fills a given slot with a given value indirectly (i.e. writing an
	** inverse slot in response to writing its `FOR' slot). Doesn't record
	** a direct write on this slot, and doesn't look for any inverses to
	** fill (because an inverse attribute cannot be the `FOR' attribute of
	** another inverse attribute).
	** @param slot The slot you want to write into.
	** @param filler The filler (value) you want to write into `slot',
	**               which must be of the right type for `slot'.
	***********************************************************************/

	public void setSlotFillerIndirectly (Slot slot, Object filler) {
		if (filler != null)
			_frameRep.put(slot,filler);
		else
			_frameRep.remove(slot);

		if (filler instanceof Frame) {
			if (_caseProblemPart)
				((Frame)filler).markCaseProblemPart();
		}
		else if (filler instanceof Vector) {
			Enumeration fillers= ((Vector)filler).elements();
			while (fillers.hasMoreElements()) {
				Object nextFiller= fillers.nextElement();
				if (nextFiller instanceof Frame) {
					if (_caseProblemPart)
						((Frame)nextFiller).
							markCaseProblemPart();
				}
				else break;
			}
		}
	}

	/**********************************************************************/
	// Frame.setSlotFiller : (String x Object) -> void
	/***********************************************************************
	** Fills a given slot with a given value.
	** @param slot The name of the slot you want to write into.
	** @param filler The filler (value) you want to write into the slot,
	**               which must be of the right type for the slot.
	** @exception InvalidSlot If the frame does not have that slot
	**                        (as specified in its frame class).
	***********************************************************************/
 
	public void setSlotFiller (String slotName, Object filler)
	                          throws InvalidSlot
	{
		setSlotFiller(getFrameClass().getSlot(slotName),filler);
	}

	/**********************************************************************/
	// Frame.isSlotSetDirectly : Slot -> boolean
	/***********************************************************************
	** Returns true just in case a given slot in this frame was written
	** directly (i.e. not just filled in automatically as an inverse slot).
	** @param slot A slot in this frame (presumably inverse).
	** @returns Whether `slot' has been written directly in this frame.
	***********************************************************************/

	public boolean isSlotSetDirectly (Slot slot) {
		return _slotsSetDirectly.get(slot) != null;
	}

	/**********************************************************************/
	// Frame.getSlots : void -> Enumeration
	/***********************************************************************
	** Returns an enumeration of the filled slots in this frame.
	** @returns An enumeration of the filled slots in this frame.
	***********************************************************************/

	public Enumeration getSlots () {
		return _frameRep.keys();
	}

	/**********************************************************************/
	// Frame.markCaseProblemPart : void -> void
	/***********************************************************************
	** Marks this frame as being in the problem part of a case, recursing
	** for all frames in the slot fillers of this frame.
	***********************************************************************/

	public void markCaseProblemPart () {
		if (! _caseProblemPart) {
			_caseProblemPart= true;
		
			Enumeration filledSlots= _frameRep.keys();
			while (filledSlots.hasMoreElements()) {
				Slot slot= (Slot)filledSlots.nextElement();
				processSlotFiller(getSlotFiller(slot));
			}
		}
	}
	private void processSlotFiller (Object filler) {
		if (filler instanceof Frame)
			((Frame)filler).markCaseProblemPart();
		else if (filler instanceof Vector) {
			Enumeration elements= ((Vector)filler).elements();
			while (elements.hasMoreElements()) {
				processSlotFiller(elements.nextElement());
			}
		}
	}

	/**********************************************************************/
	// Frame.isCaseProblemPart : void -> boolean
	/***********************************************************************
	** Returns true just in case this frame is in the `problem' part of a
	** case.
	** @returns Whether this frame is in the `problem' part of a case.
	***********************************************************************/

	public boolean isCaseProblemPart () {
		return _caseProblemPart;
	}

	/**********************************************************************/
	// Frame.toString : void -> String
	/***********************************************************************
	** Returns a string representing this frame. The format is
	** <tt><i>frame-class</i>#<i>number</i></tt>.
	** @returns The string representation of this frame.
	***********************************************************************/

	public String toString () {
		return getFrameClass() + "#" + _number;
	}

	/**********************************************************************/
	// Frame.print : PrintStream -> void
	/***********************************************************************
	** Prints the frame in human-readable text form. Any references are just
	** printed as they are.
	** @param ps A stream to print the frame to.
	***********************************************************************/

	public void print (PrintStream ps) {
                Slot slot;

		generateFIL(new PrintWriter(ps,true),false);
		ps.println();
	}
	
	/**********************************************************************/
	// static Frame.generateFIL : void -> void
	/***********************************************************************
 	** Writes all frames in the system to an FIL file, called
	** <TT><<I>frame-class</I>>.fil</TT>.
	** @param HTML If true, file will be augmented with HTML tags.
	***********************************************************************/

	public static void generateFIL (boolean HTML) {
		System.out.println();

		Enumeration frameClasses= FrameClass.getAllFrameClasses();
		while (frameClasses.hasMoreElements()) {
			FrameClass fc= (FrameClass)frameClasses.nextElement();
			try {
				if (((Logical)fc.getClassFrame().getSlotFiller
				("is_abstract")).value == Logical.FALSE)
					fc.generateFILforInstances(HTML);
			} catch (InvalidSlot x) {
				System.err.println("Something has gone badly "
				+ "wrong for InvalidSlot to be thrown in static"
				+ " Frame.generateFIL!");
				System.exit(1);
			}
		}
	}

	/**********************************************************************/
	// Frame.generateFIL : PrintWriter x boolean -> void
	/***********************************************************************
	** Writes the frame into an FIL file. Any references are given
	** corresponding hypertext links if the HTML flag is true.
	** @param pw A file to generate the HTML into (with automatic line
	**  flushing enabled).
	** @param HTML If true, file will be augmented with HTML tags.
	***********************************************************************/

	public void generateFIL (PrintWriter pw, boolean HTML) {
		Slot slot;
		Object filler;

		////////////
		// Header //
		////////////
		if (HTML) pw.println("<A NAME=\"" + _number +"\"></A>");
		if (HTML) pw.println("<PRE>");
		if (HTML && _caseProblemPart) pw.println("<I>");
		pw.print("INSTANCE ");
		if (HTML) pw.print("<A HREF=\"entity_type.html#"
		+ getFrameClass().getClassFrame().getNumber() + "\">");
		pw.print(getFrameClass());
		if (HTML) pw.print("</A>");
		pw.println("#" + _number + ";");

		///////////
		// Slots //
		///////////
                Enumeration slots= getFrameClass().getSlots().elements();
                while (slots.hasMoreElements()) {
                       	slot= (Slot)slots.nextElement();
                       	filler= getSlotFiller(slot);
                       	if (filler != null) {
                        	pw.print("\t");
                        	if (HTML) pw.print("<A HREF=\""
                        	+ slot.getSlotFrame().getFrameClass() + ".html#"
                        	+ slot.getSlotFrame().getNumber() + "\">");
                        	pw.print(slot.getName());
                        	if (HTML) pw.print("</A>");
                        	pw.print(" = " + Values.FILise(filler,HTML)
                        	          + ";");
                        	if (! isSlotSetDirectly(slot))
                        		pw.print
                        		   (" <EM>(* set indirectly *)</EM>");
                        	pw.println();
                        }
		}		
		pw.println("END_INSTANCE;");
		if (HTML && _caseProblemPart) pw.println("</I>");
		if (HTML) pw.println("</PRE>");
	}
}

//////////////////////////
// Subclasses of Frame: //
// - DefinedTypeFrame   //
// - EntityFrame        //
// - ExplicitAttrFrame  //
// - InverseAttrFrame   //
// - DerivedAttrFrame   //
// - DomainRuleFrame    //
// - UniqueRuleFrame    //
// - LiteralFrame       //
// - BuiltInFrame       //
// - UnaryOpFrame       //
// - BinaryOpFrame      //
// - QueryFrame         //
// - BuiltInFnCallFrame //
// - UserFnCallFrame    //
// - QualifierFrame     //
// - AggInitFrame       //
// - AggInitElemFrame   //
// - AggTypeFrame       //
// - OneofFrame         //
// - AndFrame           //
// - AndorFrame         //
//////////////////////////
