     ////////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM    ///
   // Written by Andrew Broad                ////
  //                                        /////
 // HashtableStack.java (1998:10:27 11:24) //////
/////////////////////////////////////////////////
// Methods of HashtableStack:             ///////
// - HashtableStack constructor           //////
// - push                                 /////
// - pop                                  ////
// - store                                ///
// - lookup                               //
///////////////////////////////////////////

import java.io.*;
import java.util.*;

/*******************************************************************************
** The class `HashtableStack' represents a stack of hash tables. It is used for
** storing frame associations and variable bindings with a stack-based
** retraction mechanism.
*******************************************************************************/

public class HashtableStack implements Serializable {
	private Vector _stack/*Hashtable*/;

	/**********************************************************************/
	// HashtableStack constructor : void ->
	/***********************************************************************
	** Constructs a new hash table stack with one hash table in it
	** initially.
	***********************************************************************/

	public HashtableStack () {
		_stack= new Vector();
		_stack.addElement(new Hashtable());
	}

	/**********************************************************************/
	// HashtableStack.push : void -> void
	/***********************************************************************
	** Pushes a new hash table onto this stack. All subsequent calls to
	** store will put an object in this hash table until it is popped.
	***********************************************************************/

	public void push () {
		_stack.addElement(new Hashtable());
	}

	/**********************************************************************/
	// HashtableStack.pop : void -> void
	/***********************************************************************
	** Pops the most recently pushed hash table off of the stack, causing
	** all the associations in that hash table to be retracted.
	***********************************************************************/

	public void pop () {
		int end;
		if ((end= _stack.size() - 1) == 0) {
			System.err.println("HashtableStack error - tried to pop"
			+ " the one remaining hash table off of the stack!");
			System.exit(1);
		}
		_stack.removeElementAt(end);
	}

	/**********************************************************************/
	// HashtableStack.store : (Object x Object) -> void
	/***********************************************************************
	** Stores a given (key,value) pair in the hash table at the top of the
	** stack.
	***********************************************************************/

	public void store (Object key, Object value) {
		((Hashtable)_stack.elementAt(_stack.size() - 1)).put(key,value);
	}

	/**********************************************************************/
	// HashtableStack.lookup : Object -> Object
	/***********************************************************************
	** Looks up the value with the given key across all of the hash tables
	** in the stack.
	***********************************************************************/

	public Object lookup (Object key) {
		for (int i= 0; i < _stack.size(); i++) {
			Object val;
			if ((val= ((Hashtable)_stack.elementAt(i)).
			    get(key)) != null)
				return val;
		}
		return null;
	}

	/**********************************************************************/
	// Case.showMaps : void -> void
	/***********************************************************************
	** Shows the variable bindings and frame associations for this case
	** (when it has been matched) - helpful for debugging.
	***********************************************************************/

/*	private void showMaps () {
		System.err.println("++++++++++++++++++++++++++++++++++++++++" +
		"++++++++++++++++++++++++++++++++++++++++\n_frameMap:-");
		Enumeration f= _frameMap.keys();
		while (f.hasMoreElements()) {
			Frame k= (Frame)f.nextElement();
			System.err.println("\t" + k + "->" + _frameMap.get(k));
		}
		System.err.println("_varMap:-");
		Enumeration v= _varMap.keys();
		while (v.hasMoreElements()) {
			String k= (String)v.nextElement();
			System.err.println("\t" + k + " = " + _varMap.get(k));
		}
		System.err.println("++++++++++++++++++++++++++++++++++++++++" +
		"++++++++++++++++++++++++++++++++++++++++");
	}
*/}