     /////////////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM          //
   // Written by Andrew Broad                      ///
  //                                              ////
 // UnorderedVector.java (1998:10:27 11:38)      /////
//////////////////////////////////////////////////////
// Methods of UnorderedVector:                   /////
// - UnorderedVector constructor()               /////
// - UnorderedVector constructor(Object[])       ////
// - UnorderedVector constructor(Object[],Frame) ///
// - FILise                                      //
//////////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `UnorderedVector' is a subclass of vectors that represent sets or
** bags in the frame system.
*******************************************************************************/

public class UnorderedVector extends Vector {

	/**********************************************************************/
	// UnorderedVector constructor : void ->
	/***********************************************************************
	** Constructs an empty unordered vector.
	***********************************************************************/
 
 	public UnorderedVector () {
		super();
	}
	
	/**********************************************************************/
	// UnorderedVector constructor : Object[] ->
	/***********************************************************************
	** Constructs an unordered vector from an exjava array, converting the
	** elements to their corresponding values for the frame system.
	** @param array The array of objects to construct the vector from.
	***********************************************************************/
 
 	public UnorderedVector (Object[] array) {
 		super();
 		
		for (int i= 0; i < array.length; i++)
			addElement(Values.convert(array[i]));
	}

	/**********************************************************************/
	// UnorderedVector constructor : (Object[] x Frame) ->
	/***********************************************************************
	** Constructs an unordered vector from an exjava array, converting the
	** elements to their corresponding values for the frame system. The
	** extra argument is an entity or defined type frame to be passed on to
	** Values.convert just in case the vector contains IdRefs (directly or
	** indirectly).
	** @param array The array of objects to construct the vector from.
	** @param entityorTypeFrame The entity or defined type frame to be
	**  passed to Values.convert.
	***********************************************************************/
 
 	public UnorderedVector (Object[] array, Frame entityOrTypeFrame) {
 		super();
 		
		for (int i= 0; i < array.length; i++)
			addElement(Values.convert(array[i], entityOrTypeFrame));
	}

	/**********************************************************************/
	// UnorderedVector.FILise : boolean -> String
	/***********************************************************************
	** Writes FIL for this unordered vector. Elements are enclosed in
	** curly brackets and separated by commas. Elements are FIL-ised
	** recursively.
	** @param HTML If true, FIL will be augmented with HTML tags.
	** @returns The FIL representation of this unordered vector.
	***********************************************************************/

	public String FILise (boolean HTML) {
		String agg= "{";

		for (int i= 0; i < size(); i++) {
			Object element= elementAt(i);
			if (element != null)
				agg+= Values.FILise(element,HTML);
			else
				agg+= "null";
			if (i < (size() - 1))
			agg+= ", ";
		}
		return agg + "}";
	}
}
